const IteratorSlot = Symbol('IteratorSlot');
export function wrap(func) {
    return function (...args) {
        // @ts-expect-error
        return iter(func.call(this, ...args));
    };
}
export function awrap(func) {
    return function (...args) {
        // @ts-expect-error
        return aiter(func.call(this, ...args));
    };
}
export function aiter(item) {
    if (Symbol.asyncIterator in item) {
        // @ts-expect-error
        return new HAsyncIterator(item[Symbol.asyncIterator]());
    }
    return new HAsyncIterator(item);
}
export class HAsyncIterator {
    constructor(iterator) {
        this[IteratorSlot] = iterator !== null && iterator !== void 0 ? iterator : this;
    }
    next(val) {
        return this[IteratorSlot].next(val);
    }
    throw(val) {
        var _a, _b, _c;
        return (_c = (_b = (_a = this[IteratorSlot]).throw) === null || _b === void 0 ? void 0 : _b.call(_a, val)) !== null && _c !== void 0 ? _c : Promise.resolve({ value: undefined, done: true });
    }
    return(val) {
        var _a, _b, _c;
        return (_c = (_b = (_a = this[IteratorSlot]).return) === null || _b === void 0 ? void 0 : _b.call(_a, val)) !== null && _c !== void 0 ? _c : Promise.resolve({ value: undefined, done: true });
    }
    /** Map each value of iterator to another value via {callback}. */
    map(callback) {
        return new HAsyncIterator(HAsyncIterator.map.call(this, callback));
    }
    static async *map(callback) {
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = await callback(value.value);
            const next_value = yield real_value;
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Each value is given through {callback}, return `true` if value is needed into returned iterator. */
    filter(callback) {
        return new HAsyncIterator(HAsyncIterator.filter.call(this, callback));
    }
    static async *filter(callback) {
        const it = this;
        let value = await it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value)) {
                next_value = yield real_value;
            }
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Find a specific value that returns `true` in {callback}, and return it. Returns `undefined` otherwise. */
    async find(callback) {
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value))
                return real_value;
            value = await it.next();
        }
    }
    /** Return `true` if each value of iterator validate {callback}. */
    async every(callback) {
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (!await callback(real_value))
                return false;
            value = await it.next();
        }
        return true;
    }
    /** Return `true` if one value of iterator validate {callback}. */
    async some(callback) {
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value))
                return true;
            value = await it.next();
        }
        return false;
    }
    /** Consume iterator and collapse values inside an array. */
    async toArray(max_count = Infinity) {
        const values = [];
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (max_count <= 0)
                return values;
            values.push(real_value);
            if (max_count !== Infinity)
                max_count--;
            value = await it.next();
        }
        return values;
    }
    /** Create a new iterator that consume {limit} items, then stops. */
    take(limit) {
        return new HAsyncIterator(HAsyncIterator.take.call(this, limit));
    }
    static async *take(limit) {
        limit = Number(limit);
        if (limit < 0)
            throw new RangeError('Invalid limit.');
        const it = this;
        let value = await it.next();
        let next_value;
        let remaining = limit;
        while (!value.done) {
            if (remaining <= 0)
                return;
            const real_value = value.value;
            next_value = yield real_value;
            value = await it.next(next_value);
            remaining--;
        }
        return value.value;
    }
    /** Create a new iterator that skip {limit} items from source iterator, then yield all values. */
    drop(limit) {
        return new HAsyncIterator(HAsyncIterator.drop.call(this, limit));
    }
    static async *drop(limit) {
        limit = Number(limit);
        if (limit < 0)
            throw new RangeError('Invalid limit.');
        const it = this;
        let value = await it.next();
        let next_value;
        let remaining = limit;
        while (!value.done) {
            if (remaining > 0) {
                remaining--;
                value = await it.next(next_value);
                continue;
            }
            const real_value = value.value;
            next_value = yield real_value;
            value = await it.next(next_value);
            remaining--;
        }
        return value.value;
    }
    /** Get a pair [index, value] for each remaining value of iterable. */
    asIndexedPairs() {
        return new HAsyncIterator(HAsyncIterator.asIndexedPairs.call(this));
    }
    static async *asIndexedPairs() {
        let index = 0;
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            const next_value = yield [index, real_value];
            index++;
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Like map, but you can return a new iterator that will be flattened. */
    flatMap(mapper) {
        return new HAsyncIterator(HAsyncIterator.flatMap.call(this, mapper));
    }
    static async *flatMap(mapper) {
        if (typeof mapper !== 'function') {
            throw new TypeError('Mapper must be a function.');
        }
        const it = this;
        let value = await it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            const mapped = await mapper(real_value);
            if (Symbol.asyncIterator in mapped) {
                // @ts-ignore
                yield* mapped[Symbol.asyncIterator]();
            }
            else if (Symbol.iterator in mapped) {
                // @ts-ignore
                yield* mapped[Symbol.iterator]();
            }
            else {
                yield mapped;
            }
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Accumulate each item inside **acc** for each value **value**. */
    async reduce(reducer, initial_value) {
        let acc = initial_value;
        const it = this;
        if (acc === undefined) {
            acc = (await it.next()).value;
        }
        for await (const value of it) {
            acc = await reducer(acc, value);
        }
        return acc;
    }
    /** Iterate over each value of iterator by calling **callback** for each value. */
    async forEach(callback) {
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            await callback(real_value);
            value = await it.next();
        }
    }
    /** Join all the remaining elements of the iterator in a single string with glue {glue}. */
    async join(string) {
        let final = '';
        let first = true;
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (first) {
                first = false;
                final += real_value;
            }
            else {
                final += string + real_value;
            }
            value = await it.next();
        }
        return final;
    }
    /** End the iterator and return the number of remaining items. */
    async count() {
        let count = 0;
        const it = this;
        let value = await it.next();
        while (!value.done) {
            count++;
            value = await it.next();
        }
        return count;
    }
    /** Iterate through current iterator, then through the given iterators in the correct order. */
    chain(...iterables) {
        return new HAsyncIterator(HAsyncIterator.chain.call(this, ...iterables));
    }
    static async *chain(...iterables) {
        yield* this;
        for (const it of iterables) {
            yield* it;
        }
    }
    /** Iterate through multiple iterators together. */
    zip(...others) {
        return new HAsyncIterator(HAsyncIterator.zip.call(this, ...others));
    }
    static async *zip(...others) {
        const it_array = [this, ...others].map((e) => Symbol.asyncIterator in e ? e[Symbol.asyncIterator]() : e);
        let values = await Promise.all(it_array.map(e => e.next()));
        while (values.every(e => !e.done)) {
            yield values.map(e => e.value);
            values = await Promise.all(it_array.map(e => e.next()));
        }
    }
    /** Continue iterator until {callback} return a falsy value. */
    takeWhile(callback) {
        return new HAsyncIterator(HAsyncIterator.takeWhile.call(this, callback));
    }
    static async *takeWhile(callback) {
        const it = this;
        let value = await it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value)) {
                next_value = yield real_value;
            }
            else {
                return;
            }
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Skip elements until {callback} return a truthy value. */
    dropWhile(callback) {
        return new HAsyncIterator(HAsyncIterator.dropWhile.call(this, callback));
    }
    static async *dropWhile(callback) {
        const it = this;
        let value = await it.next();
        let next_value;
        let finished = false;
        while (!value.done) {
            const real_value = value.value;
            if (!finished && await callback(real_value)) {
                value = await it.next(next_value);
                continue;
            }
            finished = true;
            next_value = yield real_value;
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Continue iterator until `null` or `undefined` is encountered. */
    fuse() {
        return new HAsyncIterator(HAsyncIterator.fuse.call(this));
    }
    static async *fuse() {
        const it = this;
        let value = await it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (real_value !== undefined && real_value !== null) {
                next_value = yield real_value;
            }
            else {
                return;
            }
            value = await it.next(next_value);
        }
        return value.value;
    }
    /** Partition {true} elements to first array, {false} elements to second one. */
    async partition(callback) {
        const partition1 = [], partition2 = [];
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value))
                partition1.push(real_value);
            else
                partition2.push(real_value);
            value = await it.next();
        }
        return [partition1, partition2];
    }
    /** Find the iterator index of the first element that returns a truthy value, -1 otherwise. */
    async findIndex(callback) {
        const it = this;
        let value = await it.next();
        let i = 0;
        while (!value.done) {
            const real_value = value.value;
            if (await callback(real_value))
                return i;
            value = await it.next();
            i++;
        }
        return -1;
    }
    /** Only works if it. is a number iterator. Returns the maximum of iterator. */
    async max() {
        let max = -Infinity;
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (max < real_value)
                max = real_value;
            value = await it.next();
        }
        return max;
    }
    /** Only works if it. is a number iterator. Returns the minimum of iterator. */
    async min() {
        let min = Infinity;
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            if (min > real_value)
                min = real_value;
            value = await it.next();
        }
        return min;
    }
    /** When iterator ends, go back to the first item then loop. Indefinitively. */
    cycle() {
        return new HAsyncIterator(HAsyncIterator.cycle.call(this));
    }
    static async *cycle() {
        const values = [];
        const it = this;
        let value = await it.next();
        while (!value.done) {
            const real_value = value.value;
            values.push(real_value);
            const next_value = yield real_value;
            value = await it.next(next_value);
        }
        while (true) {
            for (const value of values) {
                yield value;
            }
        }
    }
    [Symbol.asyncIterator]() {
        return this;
    }
}
export function iter(item) {
    if (Symbol.iterator in item) {
        // @ts-expect-error
        return new HIterator(item[Symbol.iterator]());
    }
    return new HIterator(item);
}
export class HIterator {
    constructor(iterator) {
        this[IteratorSlot] = iterator !== null && iterator !== void 0 ? iterator : this;
    }
    next(val) {
        return this[IteratorSlot].next(val);
    }
    throw(val) {
        var _a, _b, _c;
        return (_c = (_b = (_a = this[IteratorSlot]).throw) === null || _b === void 0 ? void 0 : _b.call(_a, val)) !== null && _c !== void 0 ? _c : { value: undefined, done: true };
    }
    return(val) {
        var _a, _b, _c;
        return (_c = (_b = (_a = this[IteratorSlot]).return) === null || _b === void 0 ? void 0 : _b.call(_a, val)) !== null && _c !== void 0 ? _c : { value: undefined, done: true };
    }
    /** Map each value of iterator to another value via {callback}. */
    map(callback) {
        return new HIterator(HIterator.map.call(this, callback));
    }
    static *map(callback) {
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = callback(value.value);
            const next_value = yield real_value;
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Each value is given through {callback}, return `true` if value is needed into returned iterator. */
    filter(callback) {
        return new HIterator(HIterator.filter.call(this, callback));
    }
    static *filter(callback) {
        const it = this;
        let value = it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value)) {
                next_value = yield real_value;
                value = it.next(next_value);
            }
            else {
                value = it.next(next_value);
            }
        }
        return value.value;
    }
    /** Find a specific value that returns `true` in {callback}, and return it. Returns `undefined` otherwise. */
    find(callback) {
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value))
                return real_value;
            value = it.next();
        }
    }
    /** Return `true` if each value of iterator validate {callback}. */
    every(callback) {
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (!callback(real_value))
                return false;
            value = it.next();
        }
        return true;
    }
    /** Return `true` if one value of iterator validate {callback}. */
    some(callback) {
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value))
                return true;
            value = it.next();
        }
        return false;
    }
    /** Consume iterator and collapse values inside an array. */
    toArray(max_count = Infinity) {
        const values = [];
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (max_count <= 0)
                return values;
            values.push(real_value);
            if (max_count !== Infinity)
                max_count--;
            value = it.next();
        }
        return values;
    }
    /** Create a new iterator that consume {limit} items, then stops. */
    take(limit) {
        limit = Number(limit);
        if (limit < 0)
            throw new RangeError('Invalid limit.');
        return new HIterator(HIterator.take.call(this, limit));
    }
    static *take(limit) {
        const it = this;
        let value = it.next();
        let remaining = limit;
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (remaining <= 0)
                return;
            next_value = yield real_value;
            value = it.next(next_value);
            remaining--;
        }
        return value.value;
    }
    /** Create a new iterator that skip {limit} items from source iterator, then yield all values. */
    drop(limit) {
        limit = Number(limit);
        if (limit < 0)
            throw new RangeError('Invalid limit.');
        return new HIterator(HIterator.drop.call(this, limit));
    }
    static *drop(limit) {
        const it = this;
        let value = it.next();
        let remaining = limit;
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (remaining > 0) {
                value = it.next(next_value);
                remaining--;
                continue;
            }
            next_value = yield real_value;
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Get a pair [index, value] for each remaining value of iterable. */
    asIndexedPairs() {
        return new HIterator(HIterator.asIndexedPairs.call(this));
    }
    static *asIndexedPairs() {
        const it = this;
        let value = it.next();
        let index = 0;
        while (!value.done) {
            const real_value = value.value;
            const next_value = yield [index, real_value];
            ;
            value = it.next(next_value);
            index++;
        }
        return value.value;
    }
    /** Like map, but you can return a new iterator that will be flattened. */
    flatMap(mapper) {
        if (typeof mapper !== 'function') {
            throw new TypeError('Mapper must be a function.');
        }
        return new HIterator(HIterator.flatMap.call(this, mapper));
    }
    static *flatMap(mapper) {
        const it = this;
        let value = it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            const mapped = mapper(real_value);
            if (Symbol.iterator in mapped) {
                // @ts-ignore
                next_value = yield* mapped[Symbol.iterator]();
            }
            else {
                // @ts-ignore
                next_value = yield mapped;
            }
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Accumulate each item inside **acc** for each value **value**. */
    reduce(reducer, initial_value) {
        let acc = initial_value;
        const it = this;
        if (acc === undefined) {
            acc = it.next().value;
        }
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            acc = reducer(acc, real_value);
            value = it.next();
        }
        return acc;
    }
    /** Iterate over each value of iterator by calling **callback** for each value. */
    forEach(callback) {
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            callback(real_value);
            value = it.next();
        }
    }
    /** End the iterator and return the number of remaining items. */
    count() {
        let count = 0;
        const it = this;
        let value = it.next();
        while (!value.done) {
            count++;
            value = it.next();
        }
        return count;
    }
    /** Join all the remaining elements of the iterator in a single string with glue {glue}. */
    join(string) {
        let final = '';
        let first = true;
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (first) {
                first = false;
                final += real_value;
            }
            else {
                final += string + real_value;
            }
            value = it.next();
        }
        return final;
    }
    /** Iterate through current iterator, then through the given iterators in the correct order. */
    chain(...iterables) {
        return new HIterator(HIterator.chain.apply(this, iterables));
    }
    static *chain(...iterables) {
        yield* this;
        for (const it of iterables) {
            yield* it;
        }
    }
    /** Iterate through multiple iterators together. */
    zip(...others) {
        return new HIterator(HIterator.zip.apply(this, others));
    }
    static *zip(...others) {
        const it_array = [this, ...others].map((e) => Symbol.iterator in e ? e[Symbol.iterator]() : e);
        let values = it_array.map(e => e.next());
        let next_value;
        while (values.every(e => !e.done)) {
            next_value = yield values.map(e => e.value);
            values = it_array.map(e => e.next(next_value));
        }
    }
    /** Continue iterator until {callback} return a falsy value. */
    takeWhile(callback) {
        return new HIterator(HIterator.takeWhile.call(this, callback));
    }
    static *takeWhile(callback) {
        const it = this;
        let value = it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value))
                next_value = yield real_value;
            else
                return;
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Skip elements until {callback} return a truthy value. */
    dropWhile(callback) {
        return new HIterator(HIterator.dropWhile.call(this, callback));
    }
    static *dropWhile(callback) {
        const it = this;
        let value = it.next();
        let next_value;
        let finished = false;
        while (!value.done) {
            const real_value = value.value;
            if (!finished && callback(real_value)) {
                value = it.next(next_value);
                continue;
            }
            finished = true;
            next_value = yield real_value;
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Continue iterator until `null` or `undefined` is encountered. */
    fuse() {
        return new HIterator(HIterator.fuse.call(this));
    }
    static *fuse() {
        const it = this;
        let value = it.next();
        let next_value;
        while (!value.done) {
            const real_value = value.value;
            if (real_value !== undefined && real_value !== null)
                next_value = yield real_value;
            else
                return;
            value = it.next(next_value);
        }
        return value.value;
    }
    /** Partition {true} elements to first array, {false} elements to second one. */
    partition(callback) {
        const partition1 = [], partition2 = [];
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value))
                partition1.push(real_value);
            else
                partition2.push(real_value);
            value = it.next();
        }
        return [partition1, partition2];
    }
    /** Find the iterator index of the first element that returns a truthy value, -1 otherwise. */
    findIndex(callback) {
        const it = this;
        let i = 0;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (callback(real_value))
                return i;
            value = it.next();
            i++;
        }
        return -1;
    }
    /** Only works if it. is a number iterator. Returns the maximum of iterator. */
    max() {
        let max = -Infinity;
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (max < real_value)
                max = real_value;
            value = it.next();
        }
        return max;
    }
    /** Only works if it. is a number iterator. Returns the minimum of iterator. */
    min() {
        let min = Infinity;
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            if (min > real_value)
                min = real_value;
            value = it.next();
        }
        return min;
    }
    /** When iterator ends, go back to the first item then loop. Indefinitively. */
    cycle() {
        return new HIterator(HIterator.cycle.call(this));
    }
    static *cycle() {
        const values = [];
        const it = this;
        let value = it.next();
        while (!value.done) {
            const real_value = value.value;
            values.push(real_value);
            const next_value = yield real_value;
            value = it.next(next_value);
        }
        while (true) {
            yield* values;
        }
    }
    [Symbol.iterator]() {
        return this;
    }
}
